Document: ibus-table
Title: Debian ibus-table Manual
Author: <insert document author here>
Abstract: This manual describes what ibus-table is
 and how it can be used to
 manage online manuals on Debian systems.
Section: unknown

Format: debiandoc-sgml
Files: /usr/share/doc/ibus-table/ibus-table.sgml.gz

Format: postscript
Files: /usr/share/doc/ibus-table/ibus-table.ps.gz

Format: text
Files: /usr/share/doc/ibus-table/ibus-table.text.gz

Format: HTML
Index: /usr/share/doc/ibus-table/html/index.html
Files: /usr/share/doc/ibus-table/html/*.html

  
